package edu.prlab.tyler.iotgateway.gateway.services;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import com.google.common.collect.Streams;
import edu.prlab.tyler.iotgateway.cloud.pojo.privacy.PrivacyChoice;
import edu.prlab.tyler.iotgateway.cloud.pojo.privacy.PrivacyPolicy;
import edu.prlab.tyler.iotgateway.gateway.model.PrivacyChoiceResponse;
import edu.prlab.tyler.iotgateway.gateway.model.PrivacyPolicyReportResponse;
import edu.prlab.tyler.iotgateway.gateway.pojo.PrivacyChoiceIndex;
import edu.prlab.tyler.iotgateway.gateway.repositories.PrivacyChoiceIndexRepository;
import edu.prlab.tyler.iotgateway.gateway.repositories.UserRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

@Service
public class PrivacyService {
    private PrivacyChoiceIndexRepository privacyChoiceIndexRepository;
    private UserRepository userRepository;
    private RemoteService remoteService;

    @Autowired
    public PrivacyService(PrivacyChoiceIndexRepository privacyChoiceIndexRepository, RemoteService remoteService,
                          UserRepository userRepository) {
        this.privacyChoiceIndexRepository = privacyChoiceIndexRepository;
        this.userRepository = userRepository;
        this.remoteService = remoteService;
    }

    // TODO 如果此User有表達過則回傳表達隱私政策與其表達，
    public Optional<PrivacyPolicyReportResponse> getRelatedPrivacyPolicyByUdn(String udn, String account) {
        return remoteService.readRelatedPrivacyPolicyReport(udn)
                .flatMap(report -> Optional.ofNullable(this.findChoicesByPolicies(report.getPolicies(), udn, account))
                        .map(this::choicesToMap)
                        .map(choices -> PrivacyPolicyReportResponse.responseBuilder()
                                .id(report.getId())
                                .version(report.getVersion())
                                .description(report.getDescription())
                                .device(report.getDevice())
                                .policies(report.getPolicies())
                                .choices(choices)
                                .build()));
    }

    private List<Boolean> findChoicesByPolicies(List<PrivacyPolicy> privacyPolicies, String udn, String account) {
        return privacyPolicies.stream()
                .map(policy -> this.findPrivacyChoiceByPrivacyPolicy(policy, udn, account))
                .map(choices -> (Iterables.isEmpty(choices)) ? Boolean.valueOf(false) : Iterables.getLast(choices).isAccepted())
                .collect(Collectors.toList());
    }

    private HashMap<Integer, Boolean> choicesToMap(List<Boolean> choices) {
        return Streams.mapWithIndex(choices.stream(),
                (choice, index) -> Maps.newHashMap(ImmutableMap.of(Long.valueOf(index).intValue(), choice)))
                .collect(HashMap::new, HashMap::putAll, HashMap::putAll);
    }

    private Iterable<PrivacyChoice> findPrivacyChoiceByPrivacyPolicy(PrivacyPolicy privacyPolicy, String udn, String account) {
        return () -> StreamSupport.stream(remoteService.readPrivacyChoices().spliterator(), false)
                .filter(choice -> choice.getPrivacyContent().getPolicy().getId().equals(privacyPolicy.getId()) &
                        choice.getPrivacyContent().getUser().getAccount().equals(account) &
                        choice.getPrivacyContent().getDevice().getUdn().equals(udn))
                .collect(Collectors.toList()).iterator();
    }

    public Optional<PrivacyChoiceResponse> setPrivacyChoice(PrivacyChoice privacyChoice) {
        return Optional.of(privacyChoice)
                .flatMap(choice -> remoteService.setPrivacyChoice(choice))
                .flatMap(choice -> this.addPrivacyChoiceIndex(PrivacyChoiceIndex.builder()
                        .id(choice.getId())
                        .localDateTime(LocalDateTime.now())
                        .build())
                        .map(choiceIndex -> PrivacyChoiceResponse.builder()
                                .id(choice.getId())
                                .localDateTime(choiceIndex.getLocalDateTime())
                                .privacyChoice(choice)
                                .build()));
    }

    private Optional<PrivacyChoiceIndex> addPrivacyChoiceIndex(PrivacyChoiceIndex choiceIndex) {
        return Optional.ofNullable(privacyChoiceIndexRepository.save(choiceIndex));
    }

//    public Optional<Iterable<PrivacyChoice>> getPrivacyPolicyChoices() {
//        return Optional.ofNullable(template.exchange(cloudAddress + "/choice", HttpMethod.GET, null,
//                new ParameterizedTypeReference<Iterable<PrivacyChoice>>() { }).getBody());
//    }

    public Iterable<PrivacyChoiceResponse> getPrivacyPolicyChoicesByAccount(String account) {
        return () -> StreamSupport.stream(remoteService.readPrivacyChoices().spliterator(), false)
                .filter(choice -> choice.getPrivacyContent().getUser().getAccount().equals(account))
                .map(choice -> privacyChoiceIndexRepository.findById(choice.getId())
                        .map(choiceIndex -> PrivacyChoiceResponse.builder()
                                .id(choiceIndex.getId())
                                .localDateTime(choiceIndex.getLocalDateTime())
                                .privacyChoice(choice)
                                .build())
                        .orElseGet(() -> new PrivacyChoiceResponse()))
                .collect(Collectors.toList())
                .iterator();
    }
}
